# Copyright (c) 2013, University of Bristol and University of Southampton.

from EStat.Templating import Template

class TemplateMVmixed2LevelMHordCC(Template):
    'Fits 2-level multivariate mixed response models (Normal, binomial, ordered, or unordered), allowing for missing data and a choice of uniform or Wishart priors. Create custom C code.'

    __version__ = '1.0.0'

    tags = [ 'Model', '2-Level', 'Normal', 'Binomial', 'Ordered multinomial', 'Unordered multinomial', 'Multivariate response']
    engines = ['CustomC']

    inputs = '''
y = DataMatrix('Responses: ')
L2ID = IDVector('Level 2 ID: ')

storeresid = Boolean('Store residuals?')

for i in range(0, len(y)):
    context['bin'+str(i+1)] = Text('Response Type for response ' + y[i] + ': ', ['Normal', 'Binomial', 'Ordered', 'Unordered'])
    if(context['bin'+str(i+1)] == 'Ordered'):
        context['ncats'+y[i]] = Integer('Number of categories for ' + y[i] + ': ')
    if(context['bin'+str(i+1)] == 'Unordered'):
        context['ncats'+y[i]] = Integer('Number of categories for ' + y[i] + ': ')
        for j in range(0, context['ncats'+y[i]]-1):
            context['x'+str(i+1)+'_'+str(j)] = DataMatrix('Explanatory variables for category ' + str(j) + ': ', allow_cat = True)
    else:
        context['x'+str(i+1)] = DataMatrix('Explanatory variables for response ' + y[i] + ': ', allow_cat = True)

lenbeta = 0
n = 0
uind = 0
for i in range(0, len(y)):
    if(context['bin'+str(i+1)] == 'Normal'):
       context['miss'+y[i]] = ParamVector(monitor=False)
       if storeresid:
           context['u' + str(uind)] = ParamVector(parents=[L2ID], as_scalar=False)
       else:
           context['u' + str(uind)] = ParamVector(parents=[L2ID], as_scalar=False, monitor=False)
    if(context['bin'+str(i+1)] == 'Binomial'):
       context['misb'+y[i]] = ParamVector(monitor=False)
       if storeresid:
           context['u' + str(uind)] = ParamVector(parents=[L2ID], as_scalar=False)
       else:
           context['u' + str(uind)] = ParamVector(parents=[L2ID], as_scalar=False, monitor=False)
    if(context['bin'+str(i+1)] == 'Ordered'):
        context['miso'+y[i]] = ParamVector(monitor=False)
        context['alpha'+y[i]] = ParamVector()
        context['alpha'+y[i]].ncols = -(context['ncats'+y[i]] -1)
        context['alpha'+y[i]].set_value(map(str, range(0, (context['ncats'+y[i]] -1))))
        if storeresid:
            context['u' + str(uind)] = ParamVector(parents=[L2ID], as_scalar=False)
        else:
            context['u' + str(uind)] = ParamVector(parents=[L2ID], as_scalar=False, monitor=False)
    if(context['bin'+str(i+1)] == 'Unordered'):
       for j in range(0, context['ncats'+y[i]]-1):
           lenbeta += len(context['x'+str(i+1)+'_'+str(j)])
           context['misu'+str(context['ncats'+y[i]])+'_'+y[i]+'_'+ str(j)] = ParamVector(monitor=False)
           if storeresid:
               context['u' + str(uind)] = ParamVector(parents=[L2ID], as_scalar=False)
           else:
               context['u' + str(uind)] = ParamVector(parents=[L2ID], as_scalar=False, monitor=False)
           uind += 1
       context['y'+ str(i+1)] = DataMatrix(value=[])
       context['origv'+str(i+1)] = Text(value = [])
       n += context['ncats'+y[i]] - 1
    else:
       lenbeta += len(context['x'+str(i+1)])
       n += 1
       if storeresid:
           context['u' + str(uind)] = ParamVector(parents=[L2ID], as_scalar=False)
       else:
           context['u' + str(uind)] = ParamVector(parents=[L2ID], as_scalar=False, monitor=False)
       uind += 1

if n == 1:
    tau = ParamScalar()
    sigma = ParamScalar(modelled = False)
    tau_u = ParamScalar()
    sigma_u = ParamScalar(modelled = False)
else:
    omega_e = ParamVector()
    omega_e.ncols = ((n+1)*n)/2
    omega_e.select('Starting values for Level 1 variance matrix (lower triangle): ')
    omega_u = ParamMatrix(modelled = False)
    omega_u.size = n
    d_e = ParamMatrix(modelled = False)
    d_e.size = n
    d_u = ParamMatrix()
    d_u.size = n
    priors2 = Text('Level 2 Priors: ', ['Uniform', 'Wishart'])
    if priors2 == 'Wishart':
        R2 = List('R matrix: ')
        v2 = Integer('Degrees of Freedom:')
beta = ParamVector()
beta.ncols = lenbeta

deviance = ParamScalar(modelled = False)
'''

    model = '''
<% lenbeta = 0 %> \\
<% n = 0 %> \\
% for i in range(0,len(context['y'])):
% if (context['bin'+str(i+1)] == 'Unordered'):
<% n = n + context['ncats'+y[i]]-1 %> \\
% else:
<% n = n + 1 %> \\
% endif
% endfor
model {
    for (i in 1:length(${y[0]})) {
    
    % if n == 1:
        miss0[i] ~ dnorm(mu0[i], tau)
        mu0[i] <- ${mmult2(context['x1'], 'beta', 'i', lenbeta)} + u0[i]
<% lenbeta += len(context['x1']) %> \\
    % else:
        dummy[i] ~ dnormal${n}a( \\
        % for i in range(0,len(context['y'])):
% if (context['bin'+str(i+1)] == 'Normal'):
miss${y[i]}[i], \\
% endif
% if (context['bin'+str(i+1)] == 'Binomial'):
misb${y[i]}[i], \\
%endif
% if (context['bin'+str(i+1)] == 'Ordered'):
miso${y[i]}[i], \\
%endif
% if (context['bin'+str(i+1)] == 'Unordered'):
% for j in range(0,context['ncats'+y[i]]-1):
misu${context['ncats'+y[i]]}_${y[i]}_${j}[i], \\
%endfor
%endif
        % endfor
        % for i in range(0,len(context['y'])):
%if (context['bin'+str(i+1)] == 'Unordered'):
% for j in range(0,context['ncats'+y[i]]-1):
mu${i}_${j}[i], \\
%endfor
%else:
mu${i}[i], \\
%endif
        % endfor
        % for i in range(0,(((n+1)*n)/2) - 1):
d_e[${i}], \\
        % endfor
d_e[${(((n+1)*n)/2) - 1}])
<% uind = 0 %> \\
        % for i in range(0, len(context['y'])):
%if (context['bin'+str(i+1)] == 'Unordered'):
% for j in range(0,context['ncats'+y[i]]-1):
        mu${i}_${j}[i] <- ${mmult2(context['x' + str(i+1)+'_'+str(j)], 'beta', 'i', lenbeta)} + u${uind}[${L2ID}[i]]
<% uind += 1 %> \\
<% lenbeta += len(context['x' + str(i+1)+ '_'+str(j)]) %> 
%endfor
%else:
	  mu${i}[i] <- ${mmult2(context['x' + str(i+1)], 'beta', 'i', lenbeta)} + u${uind}[${L2ID}[i]]
<% uind += 1 %> \\
<% lenbeta += len(context['x' + str(i+1)]) %> 
%endif
        % endfor

        % for i in range(0,len(context['y'])):
% if (context['bin'+str(i+1)] == 'Normal'):
        miss${y[i]}[i] ~ dflat()
% endif
% if (context['bin'+str(i+1)] == 'Binomial'):
        misb${y[i]}[i] ~ dflat()
%endif
% if (context['bin'+str(i+1)] == 'Ordered'):
        miso${y[i]}[i] ~ dflat()
%endif
% if (context['bin'+str(i+1)] == 'Unordered'):
% for j in range(0,context['ncats'+y[i]]-1):
        misu${context['ncats'+y[i]]}_${y[i]}_${j}[i] ~ dflat()
%endfor
%endif
        % endfor

    % endif
    }
      
    for (j in 1:length(u0)) {
        %if len(y) == 1 :
           u0[j] ~ dnorm(0, tau_u)
        % else:
        dummy_u[j] ~ dnormal${n}a( \\
        % for i in range(0,n):
u${i}[j], \\
        % endfor
        % for i in range(0,n):
0, \\
        % endfor
        % for i in range(0,(((n+1)*n)/2) - 1):
d_u[${i}], \\
        % endfor
d_u[${(((n+1)*n)/2) - 1}])

        % for i in range(0,n):
        u${i}[j] ~ dflat()
        % endfor

    % endif

    }


    # Priors

    
    % if n == 1:
    tau ~ dgamma(0.001, 0.001)
    sigma <- 1 / sqrt(tau)
    tau_u ~ dgamma(0.001, 0.001)
    sigma_u <- 1 / sqrt(tau_u)
    % endif

    % for i in range(0, lenbeta):
    beta_${i} ~ dflat()
    % endfor   
    % for i in range(0,(((n+1)*n)/2)):
    dumomega${i} ~ ddummy(omega_e${i})
    omega_e${i} ~ dflat()
    % endfor

}       
'''        

    customccode = '''
        <%!
    def mmult2(names, var, index,count):
        out = ""
        first = True
        for name in names:
            if first == False:
                out += ' + '
            else:
                first = False
            out += 'double(' + name + '[' + index + ']) * ' + var + '_' + str(count)
            count += 1
        return out

    def PyToC(var):
        if isinstance(var, list):
            return "{" + ",".join(map(PyToC, var)) + "}"
        else:
            return str(var)    
%>

            
<% nvar = len(y) %>\\
<% n = 0 %> \\
% for i in range(0, nvar):
% if (context['bin'+str(i+1)] == 'Unordered'):
<% n += context['ncats'+y[i]]-1 %> \\
% else:
<% n += 1 %> \\
% endif
% endfor

% for var in range(0, nvar):
% if (context['bin'+str(var+1)] == 'Ordered'):
    const int ncats${y[var]} = ${context['ncats'+y[var]]};
% endif  
% if (context['bin'+str(var+1)] == 'Unordered'):
    const int ncats${y[var]} = ${context['ncats'+y[var]]};
% endif  
% endfor
    {
        // Bill's attempt to include ordered cutpoints to go in here (note y step done elsewhere)
% for i in range(0, nvar):
% if (context['bin'+str(i+1)] == 'Ordered'):
        {
            double low[] = ${PyToC([1000] * int(context['ncats'+y[i]]))};
            double upp[] = ${PyToC([-1000] * int(context['ncats'+y[i]]))};

            for(int j = 0; j < length(${y[i]}); j++) {
                if(${y[i]}[j] <= -9.999e29) {
                } else {
                    if(miso${y[i]}[j] < low[int(${y[i]}[j])])
                        low[int(${y[i]}[j])] = miso${y[i]}[j];
                    if(miso${y[i]}[j] > upp[int(${y[i]}[j])])
                        upp[int(${y[i]}[j])] = miso${y[i]}[j];
                }
            } 
            for(int j = 1; j<ncats${y[i]}-1; j++) {
                double ul,ll;            
                if(upp[j]>double(alpha${y[i]}[j-1]))
                    ll = upp[j];
                else 
                    ll = double(alpha${y[i]}[j-1]);
                if(j == (ncats${y[i]}-2))
                    ul = low[j+1];
                else if (low[j+1] < double(alpha${y[i]}[j+1]))
                    ul = low[j+1];
                else
                    ul = double(alpha${y[i]}[j+1]);
                alpha${y[i]}[j]= ll+(ul-ll)*random01();
            }
        }
% endif
% endfor
        // Bill's attempt of step for d_e
        SymMatrix va(${n});
        double llcurr;
<% count = 0 %>\\
% for i in range(0, n):
% for j in range(0, i + 1):
        va(${i}, ${j}) = omega_e[${count}];
<% count += 1 %>\\
% endfor
% endfor
        bool fail = false;
        mat_d_e = matrix_sym_inverse(va, fail);
        // Note currently using a uniform prior for variance matrix
        // Note sb contains all terms needed in MH steps
        SymMatrix sb(${n});
        for(int i = 0; i < length(${y[0]}); i++) {
<% lenbeta = 0 %>\\
<% row = 0 %>\\
<% uind = 0 %>\\
% for i in range(0, nvar):
% if (context['bin'+str(i+1)] == 'Normal'):
            double e${row} = double(miss${y[i]}[i]) - (${mmult2(context['x' + str(i+1)], 'beta', 'i', lenbeta)}) - double(u${uind}[int(${L2ID}[i])]);
<% uind += 1 %>\\
<% row += 1 %>\\
% endif
% if (context['bin'+str(i+1)] == 'Binomial'):
            double e${row} = double(misb${y[i]}[i]) - (${mmult2(context['x' + str(i+1)], 'beta', 'i', lenbeta)}) - double(u${uind}[int(${L2ID}[i])]);
<% uind += 1 %>\\
<% row += 1 %>\\
% endif
% if (context['bin'+str(i+1)] == 'Ordered'):
            double e${row} = double(miso${y[i]}[i]) - (${mmult2(context['x' + str(i+1)], 'beta', 'i', lenbeta)}) - double(u${uind}[int(${L2ID}[i])]);
<% uind += 1 %>\\
<% row += 1 %>\\
% endif
% if (context['bin'+str(i+1)] == 'Unordered'):
% for j in range(0,context['ncats'+y[i]]-1):
            double e${row} = double(misu${context['ncats'+y[i]]}_${y[i]}_${j}[i]) - (${mmult2(context['x' + str(i+1)+'_'+str(j)], 'beta', 'i', lenbeta)}) - double(u${uind}[int(${L2ID}[i])]); 
<% uind += 1 %>\\
<% row += 1 %>\\
<% lenbeta += len(context['x' + str(i + 1) + '_'+ str(j)]) %>\\
% endfor
%else :
<% lenbeta += len(context['x' + str(i + 1)]) %>\\
%endif
% endfor
% for i in range(0, n):
% for j in range(i, n):
            sb(${i}, ${j}) += e${i} * e${j};
% endfor
% endfor
        }
        // Now loop over elements of va (omega_e)
<% count = 0 %>\\
<% var1 = 0 %>\\
<% withinvar1 = 0 %>\\
% for i in range(0, n):
<% var2 = var1 %>\\
<% withinvar2 = withinvar1 %>\\
% for j in range(i, n):
        {
% if var1 == var2 and (context['bin'+str(var1+1)] != 'Normal'):
            omega_e_accept[${count}] = 44;
% else:
            llcurr = length(${y[0]})*log(matrix_determinant(mat_d_e))/2;
            for(int i1 = 0; i1 < ${n}; i1++)
                for(int j1 = 0; j1 < ${n}; j1++)
                    llcurr -= sb(i1, j1) * mat_d_e(i1, j1)/2;
            double old = va(${i}, ${j});
            va(${i}, ${j}) = dnorm(va(${i}, ${j}), 1.0/omega_e_sd[${count}]);
            mat_d_e = matrix_sym_inverse(va, fail);
            if (fail == true) {
                va(${i}, ${j}) = old;
                mat_d_e = matrix_sym_inverse(va, fail);
            } else {
                double llnew = length(${y[0]})*log(matrix_determinant(mat_d_e))/2;
                for(int i1=0;i1<${n};i1++)
                    for(int j1=0;j1<${n};j1++)
                        llnew -= sb(i1, j1) * mat_d_e(i1, j1)/2;
                int change = 1;
                if(isFinite(llnew) == 0)
                    change = 0;
                else if (llnew > llcurr)
                    ;
                else {
                    if(random01() > exp(llnew-llcurr))
                        change = 0;
                }
                if(change == 0) {
                    va(${i}, ${j}) = old;
                    mat_d_e = matrix_sym_inverse(va, fail);
                } else
                    omega_e_accept[${count}]++;
            }
% endif
<% count += 1 %>\\
        }
% if (context['bin'+str(var2+1)] != 'Unordered'):
<% var2+= 1 %>\\
% elif withinvar2 != (context['ncats'+y[var2]]-2):
<% withinvar2 += 1 %>\\
%else:
<% var2+= 1 %>\\
<% withinvar2 = 0 %>\\
% endif
% endfor
% if (context['bin'+str(var1+1)] != 'Unordered'):
<% var1+= 1 %>\\
% elif withinvar1 != (context['ncats'+y[var1]]-2):
<% withinvar1 += 1 %>\\
% else:
<% var1+= 1 %>\\
<% withinvar1 = 0 %>\\
% endif
% endfor
<% count = 0 %>\\
% for i in range(0, n):
% for j in range(0, i + 1):
                omega_e[${count}] = va(${i}, ${j});
<% count += 1 %>\\
% endfor
% endfor

        SymMatrix sbu(${n});
        for(int i = 0; i < length(u0); i++) {
% for i in range(0, n):
% for j in range(i, n):
            sbu(${i}, ${j}) += double(u${i}[i]) * double(u${j}[i]);
% endfor
% endfor
        }
% if priors2 == 'Uniform':
        if (runstate == 0) {
% for i in range(0, n):
            sbu(${i}, ${i}) += 0.0001;
% endfor
        }
% endif
% if priors2 == 'Wishart':
<%
import numpy
Rmat = numpy.empty([n, n])
count = 0 
for i in range(0, n):
    for j in range(0, i + 1):
        Rmat[i, j] = float(R2[count]) 
        Rmat[j, i] = Rmat[i, j]
        count += 1
%>

<% count = 0 %>
% for i in range(0, n):
% for j in range(0, n):
        sbu(${i}, ${j}) += ${str(Rmat[i, j] * float(v2))};
% endfor
% endfor
% endif
        matrix_sym_invinplace(sbu);
% if priors2 == 'Uniform':
        int vwu = length(u0) - ${n + 1};
% endif
% if priors2 == 'Wishart':
        int vwu = length(u0) + ${v2};
% endif
        mat_d_u = dwishart(vwu, sbu, fail);
        mat_omega_u = matrix_sym_inverse(mat_d_u, fail);

















                // Step for beta
                std::vector<int> betaind(${lenbeta});
<% betacount = 0 %>
<% row = 0 %>\\
% for i in range(0, len(y)):
% if (context['bin'+str(i+1)] == 'Unordered'):
% for j in range(0, context['ncats'+y[i]]-1):
% for k in range(0, len(context['x'+str(i+1)+'_'+str(j)])):
                betaind[${betacount}] = ${row};
<% betacount += 1 %>
% endfor
<% row += 1 %>
% endfor
% else:
% for j in range(0, len(context['x' +str(i+1)])):
                betaind[${betacount}] = ${row};
<% betacount += 1 %>
% endfor
<% row += 1 %>
% endif
% endfor

                std::vector<int> betamap(${n});
                int pos = 0;
<% ycount = 0 %>
% for i in range(0, len(y)):
% if (context['bin'+str(i+1)] == 'Unordered'):
% for j in range(0, context['ncats'+y[i]]-1):
                betamap[${ycount}] = pos;
                pos += mat_x${i + 1}_${j}.cols();
<% ycount += 1 %>
% endfor
% else:
                betamap[${ycount}] = pos;
                pos += mat_x${i + 1}.cols();
<% ycount += 1 %>
% endif
% endfor


                std::vector<double*> tmp_x;
% for i in range(0, len(y)):
% if (context['bin'+str(i+1)] == 'Unordered'):
% for j in range(0, context['ncats'+y[i]]-1):
% for k in range(0, len(context['x'+str(i+1)+'_'+str(j)])):
                tmp_x.push_back((double*)${context['x' + str(i+1)+'_'+str(j)][k]});
% endfor
% endfor
% else:
% for j in range(0, len(context['x' + str(i+1)])):
                tmp_x.push_back((double*)${context['x' + str(i+1)][j]});
% endfor
% endif
% endfor
                RectMatrix mat_x(tmp_x, length(${y[0]}));

                std::vector<double*> tmp_ymiss;
<%
pyyind = []
pyymap = []
pypos = 0
%>
% for i in range(0, len(y)):
% if (context['bin'+str(i+1)] == 'Normal'):
                tmp_ymiss.push_back(miss${y[i]});
<%
pyymap.append(pypos)
pypos += 1
pyyind.append(i)
%>
% endif
% if (context['bin'+str(i+1)] == 'Binomial'):
                tmp_ymiss.push_back(misb${y[i]});
<%
pyymap.append(pypos)
pypos += 1
pyyind.append(i)
%>
% endif
% if (context['bin'+str(i+1)] == 'Ordered'):
                tmp_ymiss.push_back(miso${y[i]});
<%
pyymap.append(pypos)
pypos += 1
pyyind.append(i)
%>
% endif
% if (context['bin'+str(i+1)] == 'Unordered'):
<%
pyymap.append(pypos)
pypos += 1
%>
% for j in range(0,context['ncats'+y[i]]-1):
                tmp_ymiss.push_back(misu${context['ncats'+y[i]]}_${y[i]}_${j});
<% pyyind.append(i)%>
% endfor
% endif
% endfor
                RectMatrix mat_ymiss(tmp_ymiss, length(${y[0]}));

                std::vector<double*> tmp_u;
% for i in range(0, n):
                tmp_u.push_back(u${i});
% endfor
                RectMatrix mat_u(tmp_u, length(u0));

                SymMatrix db(${lenbeta});
                for (int i = 0; i < length(${y[0]}); i++) {
                    for (int j1 = 0; j1 < ${lenbeta}; j1++) {
                        for (int j2 = 0; j2 < j1+1; j2++) {
                            db(j1, j2) += mat_x(i, j1) * mat_d_e(betaind[j1], betaind[j2]) * mat_x(i, j2);
                        }
                    }
                }

                RectMatrix xtdy(${lenbeta}, 1);
                for (int i = 0; i < length(${y[0]}); i++) {
                    for (int j1 = 0; j1 < ${lenbeta}; j1++) {
                        for (int j2 = 0; j2 < ${n}; j2++) {
                            xtdy(j1, 0) += mat_x(i, j1) * mat_d_e(betaind[j1], j2) * (mat_ymiss(i, j2) - mat_u(${L2ID}[i], betaind[j2]));
                        }
                    }
                }

                fail = false;
                // int err = matrix_sym_invinplace(db);
                // RectMatrix betahat = db * xtdy;
                // mat_beta = dmultnormal(betahat, db, fail);

                RectMatrix betahat = matrix_solve(db, xtdy);
                mat_beta = dmultnormalinv(betahat, db, fail);




% for i in range(0, n):
<%
mat_x_i = 'mat_x' + str(pyyind[i] + 1)
if (context['bin'+str(pyyind[i]+1)] == 'Unordered'):
    mat_x_i += '_' + str(i - pyymap[pyyind[i]])
%>
                for (int j = 0; j < mat_u${i}.rows(); j++) {
                    std::vector<double> sumx(mat_x.cols());
                    std::vector<double> sumy(mat_ymiss.cols());
                    std::vector<double> sumu(${n});

                    for (int id_k = 0; id_k < idmap_${L2ID}[j].size(); id_k++) {
                        int k = idmap_${L2ID}[j][id_k];
                        for (int m = 0; m < mat_x.cols(); m++) {
                            sumx[m] += mat_x(k, m);
                        }
                        for (int m = 0; m < mat_ymiss.cols(); m++) {
                            sumy[m] += mat_ymiss(k, m);
                        }
% for m in range(0, n):
% if i != m:
                        sumu[${m}] += mat_u${m}(${L2ID}[k], 0);
% endif
% endfor
                    }

                    double tmp = 0.0;
                    for (int k = 0; k < ${mat_x_i}.cols(); k++) {
                        tmp += mat_beta(betamap[${i}] + k, 0) * sumx[betamap[${i}] + k];
                    }
                    tmp -= sumy[${i}];
                    tmp *= mat_d_e(${i}, ${i});

% for m in range(0, n):
<%
mat_x_m = 'mat_x' + str(pyyind[m] + 1)
if (context['bin'+str(pyyind[m]+1)] == 'Unordered'):
    mat_x_m += '_' + str(m - pyymap[pyyind[m]])
%>
% if i != m:
                    {
                        double tmp2 = 0.0;
                        for (int k = 0; k < ${mat_x_m}.cols(); k++) {
                            tmp2 += mat_beta(betamap[${m}] + k, 0) * sumx[betamap[${m}] + k];
                        }
                        tmp2 -= sumy[${m}];
                        tmp2 += sumu[${m}];
                        tmp2 *= mat_d_e(${i}, ${m});
                        tmp += tmp2;
                    }
% endif
% endfor

% for m in range(0, n):
% if i != m:
                    tmp += mat_d_u(${i},${m}) * mat_u${m}(j, 0);
% endif
% endfor

                    tmp /= (mat_d_u(${i}, ${i}) + mat_d_e(${i}, ${i}) * idmap_${L2ID}[j].size());

                    mat_u${i}(j, 0) = dnorm(-tmp, mat_d_u(${i}, ${i}) + mat_d_e(${i}, ${i}) * idmap_${L2ID}[j].size());
                }
% endfor


% for i in range(0, n):

<%
mat_x_i = 'mat_x' + str(pyyind[i] + 1)
if (context['bin'+str(pyyind[i]+1)] == 'Unordered'):
    mat_x_i += '_' + str(i - pyymap[pyyind[i]])
%>

                for (int j = 0; j < mat_ymiss.rows(); j++) {
                    //if (mat_y(j, ${pyyind[i]}) <= -9.999e29) {
                    {
                        double tmp = mat_u${i}(${L2ID}[j], 0);
                        for (int k = 0; k < ${mat_x_i}.cols(); k++) {
                            tmp += mat_beta(betamap[${i}] + k, 0) * ${mat_x_i}(j, k);
                        }
                        for (int k = 0; k < mat_ymiss.cols(); k++) {
                            if (k != ${i}) {
                                tmp -= mat_d_e(${i}, k) * mat_ymiss(j, k) * (1.0 / mat_d_e(${i}, ${i}));
                            }
                        }
% for m in range(0, n):
% if i != m:
<%
mat_x_m = 'mat_x' + str(pyyind[m] + 1)
if (context['bin'+str(pyyind[m]+1)] == 'Unordered'):
    mat_x_m += '_' + str(m - pyymap[pyyind[m]])
%>
                        tmp += mat_d_e(${i}, ${m}) * mat_u${m}(${L2ID}[j], 0) * (1.0 / mat_d_e(${i}, ${i}));
                        for (int k = 0; k < ${mat_x_m}.cols(); k++) {
                            tmp += mat_d_e(${i}, ${m}) * (1.0 / mat_d_e(${i}, ${i})) * mat_beta(betamap[${m}] + k, 0) * ${mat_x_m}(j, k);
                        }
% endif
% endfor


% if (context['bin'+str(pyyind[i]+1)] == 'Normal'):
                    if (mat_y(j, ${pyyind[i]}) <= -9.999e29) {
                        mat_ymiss(j, ${i}) = dnorm(tmp, mat_d_e(${i}, ${i}));
                    }
% endif
% if (context['bin'+str(pyyind[i]+1)] == 'Binomial'):
                    if (mat_y(j, ${pyyind[i]}) <= -9.999e29) 
                        mat_ymiss(j, ${i}) = dnorm(tmp, mat_d_e(${i}, ${i}));
                    else if(mat_y(j, ${pyyind[i]}) <= 0) 
                        mat_ymiss(j, ${i}) = dtnormal(tmp,1/sqrt(mat_d_e(${i}, ${i})), 2, 0, 0);
                    else          
                        mat_ymiss(j, ${i}) = dtnormal(tmp,1/sqrt(mat_d_e(${i}, ${i})), 1, 0, 0);
% endif
% if (context['bin'+str(pyyind[i]+1)] == 'Ordered'):
                    if (mat_y(j, ${pyyind[i]}) <= -9.999e29) 
                        mat_ymiss(j, ${i}) = dnorm(tmp, mat_d_e(${i}, ${i}));
                    else if(mat_y(j, ${pyyind[i]}) == 0) 
                        mat_ymiss(j, ${i}) = dtnormal(tmp, 1/sqrt(mat_d_e(${i}, ${i})), 2, 0, 0);
                    else if(mat_y(j, ${pyyind[i]}) == ${context['ncats'+y[pyyind[i]]] - 1})
                        mat_ymiss(j, ${i}) = dtnormal(tmp, 1/sqrt(mat_d_e(${i}, ${i})), 1, alpha${y[i]}[ncats${y[pyyind[i]]}-2], 0);
                    else          
                        mat_ymiss(j, ${i}) = dtnormal(tmp, 1/sqrt(mat_d_e(${i}, ${i})), 3, alpha${y[i]}[int(mat_y(j, ${i}))-1], alpha${y[i]}[int(mat_y(j, ${i}))]);

% endif

% if (context['bin'+str(pyyind[i]+1)] == 'Unordered'):
                    if (mat_y(j, ${pyyind[i]}) <= -9.999e29) 
                        mat_ymiss(j, ${i}) = dnorm(tmp, mat_d_e(${i}, ${i}));
                    else if (mat_y(j, ${pyyind[i]}) == ${context['ncats'+y[pyyind[i]]] - 1}) 
                        mat_ymiss(j, ${i}) = dtnormal(tmp, 1/sqrt(mat_d_e(${i}, ${i})), 2, 0, 0);
                    else if (mat_y(j, ${pyyind[i]}) == ${i - pyymap[pyyind[i]]}) {
                        double limit = 0;
                        for (int k = 0; k < ${context['ncats'+y[pyyind[i]]] - 1}; k++) {
                            if ((k != ${i - pyymap[pyyind[i]]}) && (mat_ymiss(j, ${pyymap[pyyind[i]]} + k) > limit))
                                limit = mat_ymiss(j, ${pyymap[pyyind[i]]} + k); 
                        }
                        mat_ymiss(j, ${i}) = dtnormal(tmp, 1/sqrt(mat_d_e(${i}, ${i})), 1, limit, 0);
                    } else {  
                        double limit = mat_ymiss(j, ${pyymap[pyyind[i]]});
                        for (int k = 0; k < ${context['ncats'+y[pyyind[i]]] - 1}; k++) {
                            if (mat_ymiss(j, ${pyymap[pyyind[i]]} + k) > limit)
                                limit = mat_ymiss(j, ${pyymap[pyyind[i]]} + k); 
                        }
                        mat_ymiss(j, ${i}) = dtnormal(tmp, 1/sqrt(mat_d_e(${i}, ${i})), 2, 0, limit);
                    }

% endif

                    }
                }
% endfor


    }
'''

    preparedata = '''
mydata = data['datafile']

for i in range(0, len(y)):
    if (context['bin'+str(i+1)] == 'Normal'):
        context['miss'+y[i]].ncols = -1*len(mydata.variables[y[i]]['data'])
    if (context['bin'+str(i+1)] == 'Binomial'):
        context['misb'+y[i]].ncols = -1*len(mydata.variables[y[i]]['data'])
    if (context['bin'+str(i+1)] == 'Ordered'):
        context['miso'+y[i]].ncols = -1*len(mydata.variables[y[i]]['data'])
    if (context['bin'+str(i+1)] == 'Unordered'):
        for j in range(0, context['ncats'+y[i]]-1):
            context['misu'+str(context['ncats'+y[i]])+'_'+y[i]+'_'+str(j)].ncols = -1*len(mydata.variables[y[i]]['data'])
        context['origv'+str(i+1)].name.append(y[i]) 
        for var in context['origv'+str(i+1)].name:
            uniqvals = list(set(mydata.variables[var]['data'].compressed()))
            uniqvals.sort()
            uniqvals.remove(uniqvals[len(uniqvals)-1])
            for j in range(0, len(uniqvals)):
                mydata.addvariable(var + '_' + str(int(j)), data = (mydata.variables[var]['data'][:] == uniqvals[j]).astype(float))
                context['y'+str(i+1)].name.append(var + '_' + str(int(j)))
'''
